require "readline"
require_relative "printer"
require_relative "types"
require_relative "reader"

NS = {
  :+ => lambda {|x, y| x + y},
  :- => lambda {|x, y| x - y},
  :* => lambda {|x, y| x * y},
  :/ => lambda {|x, y |Integer(x/y)},
  :"=" => lambda {|x, y| x == y},
  :< => lambda {|x, y| x < y},
  :> => lambda {|x, y| x > y},
  :<= => lambda {|x, y| x <= y},
  :>= => lambda {|x, y| x >= y},
  :list => lambda {|*args| List.new args},
  :list? => lambda {|x| x.is_a? List},
  :vec => lambda {|list| Vector.new list},
  :empty? => lambda {|x| x.empty?},
  :nil? => lambda {|x| x == nil},
  :true? => lambda {|x| x == true},
  :false? => lambda {|x| x == false},
  :count => lambda {|x| x==nil ? 0 : x.count},
  :throw => lambda {|val| raise MalException.new val},
  :nth => lambda {|list, idx| idx < list.length ? list[idx] : raise("nth: index out of range.")},
  :first => lambda {|l| if l != nil and not l.empty? then l[0] else nil end},
  :rest => lambda {|l| if l != nil and not l.empty? then List.new l[1..-1] else List.new end},
  :apply => lambda {|f, *args, list| f.is_a?(Fun) ? f.fn.call(*list.prepend(*args)) : f.call(*list.prepend(*args))},
  :map => lambda {|f, l| result = List.new; l.each {|elt| result.append(f.is_a?(Fun) ? f.fn.call(elt) : f.call(elt))}; result},
  :"pr-str" => lambda {|*args| args.map{|s| pr_str(s)}.join(" ")},
  :str => lambda {|*args| args.map{|s| pr_str(s, false)}.join("").gsub("\"", "")},
  :prn => lambda {|*args| puts args.map{|s| pr_str(s)}.join(" "); return nil},
  :println => lambda {|*args| puts args.map{|s| pr_str(s, false)}.join(" "); return nil},
  :"read-string" => lambda {|str| read_str str},
  :slurp => lambda {|name| File.read(name)},
  :atom => lambda {|a| Atom.new a},
  :atom? => lambda {|a| a.class == Atom},
  :deref => lambda {|a| a.val},
  :reset! => lambda {|a, val| a.val = val; val},
  :swap! => lambda {|*a| fun = a[1].class == Fun ? a[1].fn : a[1]
                     a[0].val = fun.call(a[0].val, *a[2..-1]); a[0].val},
  :cons => lambda {|a, list| array = Array.new list; List.new(array.prepend(a))},
  :concat => lambda {|*a| if a.length > 0
                        ret = []
                        a.each do |list| ret = ret + list end
                        List.new ret
                        else
                          List.new
                        end},
  :symbol => lambda {|s| s.to_sym},
  :symbol? => lambda {|x| x.is_a? Symbol},
  :keyword => lambda {|s| s.start_with?("\u029e") ? s : "\u029e"+s},
  :keyword? => lambda {|x| x.start_with? "\u029e"},
  :vector => lambda {|*a| Vector.new a},
  :vector? => lambda {|x| x.is_a? Vector},
  :sequential? => lambda {|x| x.is_a? List or x.is_a? Vector},
  :"hash-map" => lambda {|*a| if a.length.even? then return Hash[a.each_slice(2).to_a] end},
  :map? => lambda {|x| x.is_a? Hash},
  :fn? => lambda {|*a| raise "Not implemented"},
  :string? => lambda {|*a| raise "Not implemented"},
  :number? => lambda {|*a| raise "Not implemented"},
  :assoc => lambda {|h, *a| h.merge(Hash[a.each_slice(2).to_a])},
  :dissoc => lambda {|h, *a| r=h.clone; a.each do |l|
                              if l.is_a? String then r.delete(l) else l.each {|k| r.delete(k)} end
                            end; r},
  :get => lambda {|h, k| h ? h[k] : nil},
  :contains? => lambda {|h, k| h.key? k},
  :keys => lambda {|h| List.new(h.keys)},
  :vals => lambda {|h| List.new(h.values)},
  :readline => lambda {|p| Readline.readline(p, true)},\
  :"time-ms" => lambda {|*a| raise "Not implemented"},
  :meta => lambda {|*a| raise "Not implemented"},
  :"with-meta" => lambda {|*a| raise "Not implemented"},
  :seq => lambda {|*a| raise "Not implemented"},
  :conj => lambda {|*a| raise "Not implemented"}
}
