require "readline"
require_relative "reader"
require_relative "printer"
require_relative "types"

def READ(str)
  return read_str str
end

$repl_env = {:+ => lambda {|x, y| x + y},
            :- => lambda {|x, y| x - y},
            :* => lambda {|x, y| x * y},
            :/ => lambda {|x, y| Integer(x/y)}}

def eval_ast(ast, env)
  return case ast
  when Symbol
    if env.key? ast
      env[ast]
    else
      raise "'" + ast + "' not found."
    end
  when List, Vector
    ast.class.new ast.map { |x| EVAL(x, env) }
  when Hash
    ast.map { |k, v| [k, EVAL(v, env)] }.to_h
  else
    ast
  end
end

def EVAL(ast, env)
  if ast.class != List then return eval_ast(ast, env) end
  if ast.empty? then return ast end
  result = eval_ast(ast, env)
  result[0].call *result[1..-1]
end

def PRINT(str)
  pr_str str
end

def rep(str)
  PRINT(EVAL(READ(str), $repl_env))
end


while buf = Readline.readline("user> ", true)
  begin
    puts rep buf
  rescue Exception => e
    puts "Error: #{e}" 
    puts "\t#{e.backtrace.join("\n\t")}"
  end
end
