require "readline"
require_relative "reader"
require_relative "printer"
require_relative "types"
require_relative "env"

def READ(str)
  return read_str str
end

$repl_env = Env.new
$repl_env.set(:+, lambda {|x, y| x + y})
$repl_env.set(:-, lambda {|x, y| x - y})
$repl_env.set(:*, lambda {|x, y| x * y})
$repl_env.set(:/, lambda {|x, y |Integer(x/y) })

def eval_ast(ast, env)
  return case ast
  when Symbol
    env.get(ast)
  when List, Vector
    ast.class.new ast.map { |x| EVAL(x, env) }
  when Hash
    ast.map { |k, v| [k, EVAL(v, env)] }.to_h
  else
    ast
  end
end

def EVAL(ast, env)
  if ast.class != List then return eval_ast(ast, env) end
  if ast.empty? then return ast end
  # Apply 
  case ast[0]
  when :def!
    env.set(ast[1], EVAL(ast[2], env))
  when :"let*" 
    let_env = Env.new env
    ast[1].each_slice(2) {|x| let_env.set(x[0], EVAL(x[1], let_env))}
    return EVAL(ast[2], let_env)
  else
    result = eval_ast(ast, env)
    return result[0].call *result[1..-1]
  end
end

def PRINT(str)
  pr_str str
end

def rep(str)
  PRINT(EVAL(READ(str), $repl_env))
end


while buf = Readline.readline("user> ", true)
  begin
    puts rep buf
  rescue Exception => e
    puts "Error: #{e}" 
    puts "\t#{e.backtrace.join("\n\t")}"
  end
end
